from __future__ import print_function

import numpy as np
from PIL import Image
import inspect
import re
import numpy as np
import os
import collections
import matplotlib.pyplot as plt
from scipy.ndimage.interpolation import zoom
from skimage.measure import compare_ssim
import torch
from IPython import embed
import cv2
from datetime import datetime

def datetime_str():
    now = datetime.now()
    return '%04d-%02d-%02d-%02d-%02d-%02d'%(now.year,now.month,now.day,now.hour,now.minute,now.second)

def read_text_file(in_path):
    fid = open(in_path,'r')

    vals = []
    cur_line = fid.readline()
    while(cur_line!=''):
        vals.append(float(cur_line))
        cur_line = fid.readline()

    fid.close()
    return np.array(vals)

def bootstrap(in_vec,num_samples=100,bootfunc=np.mean):
    from astropy import stats
    return stats.bootstrap(np.array(in_vec),bootnum=num_samples,bootfunc=bootfunc)

def rand_flip(input1,input2):
    if(np.random.binomial(1,.5)==1):
        return (input1,input2)
    else:
        return (input2,input1)

def l2(p0, p1, range=255.):
    return .5*np.mean((p0 / range - p1 / range)**2)

def psnr(p0, p1, peak=255.):
    return 10*np.log10(peak**2/np.mean((1.*p0-1.*p1)**2))

def dssim(p0, p1, range=255.):
    # embed()
    return (1 - compare_ssim(p0, p1, data_range=range, multichannel=True)) / 2.

def rgb2lab(in_img,mean_cent=False):
    from skimage import color
    img_lab = color.rgb2lab(in_img)
    if(mean_cent):
        img_lab[:,:,0] = img_lab[:,:,0]-50
    return img_lab

def normalize_blob(in_feat,eps=1e-10):
    norm_factor = np.sqrt(np.sum(in_feat**2,axis=1,keepdims=True))
    return in_feat/(norm_factor+eps)

def cos_sim_blob(in0,in1):
    in0_norm = normalize_blob(in0)
    in1_norm = normalize_blob(in1)
    (N,C,X,Y) = in0_norm.shape

    return np.mean(np.mean(np.sum(in0_norm*in1_norm,axis=1),axis=1),axis=1)

def normalize_tensor(in_feat,eps=1e-10):
    # norm_factor = torch.sqrt(torch.sum(in_feat**2,dim=1)).view(in_feat.size()[0],1,in_feat.size()[2],in_feat.size()[3]).repeat(1,in_feat.size()[1],1,1)
    norm_factor = torch.sqrt(torch.sum(in_feat**2,dim=1)).view(in_feat.size()[0],1,in_feat.size()[2],in_feat.size()[3])
    return in_feat/(norm_factor.expand_as(in_feat)+eps)

def cos_sim(in0,in1):
    in0_norm = normalize_tensor(in0)
    in1_norm = normalize_tensor(in1)
    N = in0.size()[0]
    X = in0.size()[2]
    Y = in0.size()[3]

    return torch.mean(torch.mean(torch.sum(in0_norm*in1_norm,dim=1).view(N,1,X,Y),dim=2).view(N,1,1,Y),dim=3).view(N)

# Converts a Tensor into a Numpy array
# |imtype|: the desired type of the conve

def tensor2np(tensor_obj):
    # change dimension of a tensor object into a numpy array
    return tensor_obj[0].cpu().float().numpy().transpose((1,2,0))

def np2tensor(np_obj):
     # change dimenion of np array into tensor array
    return torch.Tensor(np_obj[:, :, :, np.newaxis].transpose((3, 2, 0, 1)))

def tensor2tensorlab(image_tensor,to_norm=True,mc_only=False):
    # image tensor to lab tensor
    from skimage import color

    img = tensor2im(image_tensor)
    # print('img_rgb',img.flatten())
    img_lab = color.rgb2lab(img)
    # print('img_lab',img_lab.flatten())
    if(mc_only):
        img_lab[:,:,0] = img_lab[:,:,0]-50
    if(to_norm and not mc_only):
        img_lab[:,:,0] = img_lab[:,:,0]-50
        img_lab = img_lab/100.

    return np2tensor(img_lab)

def tensorlab2tensor(lab_tensor,return_inbnd=False):
    from skimage import color
    import warnings
    warnings.filterwarnings("ignore")

    lab = tensor2np(lab_tensor)*100.
    lab[:,:,0] = lab[:,:,0]+50
    # print('lab',lab)

    rgb_back = 255.*np.clip(color.lab2rgb(lab.astype('float')),0,1)
    # print('rgb',rgb_back)
    if(return_inbnd):
        # convert back to lab, see if we match
        lab_back = color.rgb2lab(rgb_back.astype('uint8'))
        # print('lab_back',lab_back)
        # print('lab==lab_back',np.isclose(lab_back,lab,atol=1.))
        # print('lab-lab_back',np.abs(lab-lab_back))
        mask = 1.*np.isclose(lab_back,lab,atol=2.)
        mask = np2tensor(np.prod(mask,axis=2)[:,:,np.newaxis])
        return (im2tensor(rgb_back),mask)
    else:
        return im2tensor(rgb_back)

def tensor2im(image_tensor, imtype=np.uint8, cent=1., factor=255./2.):
# def tensor2im(image_tensor, imtype=np.uint8, cent=1., factor=1.):
    image_numpy = image_tensor[0].cpu().float().numpy()
    image_numpy = (np.transpose(image_numpy, (1, 2, 0)) + cent) * factor
    return image_numpy.astype(imtype)

def im2tensor(image, imtype=np.uint8, cent=1., factor=255./2.):
# def im2tensor(image, imtype=np.uint8, cent=1., factor=1.):
    return torch.Tensor((image / factor - cent)
                        [:, :, :, np.newaxis].transpose((3, 2, 0, 1)))

def tensor2vec(vector_tensor):
    return vector_tensor.data.cpu().numpy()[:, :, 0, 0]

def diagnose_network(net, name='network'):
    mean = 0.0
    count = 0
    for param in net.parameters():
        if param.grad is not None:
            mean += torch.mean(torch.abs(param.grad.data))
            count += 1
    if count > 0:
        mean = mean / count
    print(name)
    print(mean)

def grab_patch(img_in, P, yy, xx):
    return img_in[yy:yy+P,xx:xx+P,:]

def load_image(path):
    if(path[-3:] == 'dng'):
        import rawpy
        with rawpy.imread(path) as raw:
            img = raw.postprocess()
        # img = plt.imread(path)
    elif(path[-3:]=='bmp' or path[-3:]=='jpg' or path[-3:]=='png'):
        # import cv2
        img = cv2.imread(path, cv2.IMREAD_UNCHANGED)
        if img.ndim == 2:
            img = np.expand_dims(img, axis = 2)
            img = np.repeat(img, 3, axis = 2 )
        # print(img.shape)
        return img[:,:,::-1]
    else:
        img = (255*plt.imread(path)[:,:,:3]).astype('uint8')

    return img


def resize_image(img, max_size=256):
    [Y, X] = img.shape[:2]

    # resize
    max_dim = max([Y, X])
    zoom_factor = 1. * max_size / max_dim
    img = zoom(img, [zoom_factor, zoom_factor, 1])

    return img

def resize_image_zoom(img, zoom_factor=1., order=3):
    if(zoom_factor==1):
        return img
    else:
        return zoom(img, [zoom_factor, zoom_factor, 1], order=order)

def save_image(image_numpy, image_path, ):
    image_pil = Image.fromarray(image_numpy)
    image_pil.save(image_path)


def prep_display_image(img, dtype='uint8'):
    if(dtype == 'uint8'):
        return np.clip(img, 0, 255).astype('uint8')
    else:
        return np.clip(img, 0, 1.)


def info(object, spacing=10, collapse=1):
    """Print methods and doc strings.
    Takes module, class, list, dictionary, or string."""
    methodList = [
        e for e in dir(object) if isinstance(
            getattr(
                object,
                e),
            collections.Callable)]
    processFunc = collapse and (lambda s: " ".join(s.split())) or (lambda s: s)
    print("\n".join(["%s %s" %
                     (method.ljust(spacing),
                      processFunc(str(getattr(object, method).__doc__)))
                     for method in methodList]))


def varname(p):
    for line in inspect.getframeinfo(inspect.currentframe().f_back)[3]:
        m = re.search(r'\bvarname\s*\(\s*([A-Za-z_][A-Za-z0-9_]*)\s*\)', line)
        if m:
            return m.group(1)


def print_numpy(x, val=True, shp=False):
    x = x.astype(np.float64)
    if shp:
        print('shape,', x.shape)
    if val:
        x = x.flatten()
        print(
            'mean = %3.3f, min = %3.3f, max = %3.3f, median = %3.3f, std=%3.3f' %
            (np.mean(x), np.min(x), np.max(x), np.median(x), np.std(x)))


def mkdirs(paths):
    if isinstance(paths, list) and not isinstance(paths, str):
        for path in paths:
            mkdir(path)
    else:
        mkdir(paths)


def mkdir(path):
    if not os.path.exists(path):
        os.makedirs(path)


def rgb2lab(input):
    from skimage import color
    return color.rgb2lab(input / 255.)


def montage(
    imgs,
    PAD=5,
    RATIO=16 / 9.,
    EXTRA_PAD=(
        False,
        False),
        MM=-1,
        NN=-1,
        primeDir=0,
        verbose=False,
        returnGridPos=False,
        backClr=np.array(
            (0,
             0,
             0))):
    # INPUTS
    #   imgs        YxXxMxN or YxXxN
    #   PAD         scalar              number of pixels in between
    #   RATIO       scalar              target ratio of cols/rows
    #   MM          scalar              # rows, if specified, overrides RATIO
    #   NN          scalar              # columns, if specified, overrides RATIO
    #   primeDir    scalar              0 for top-to-bottom, 1 for left-to-right
    # OUTPUTS
    #   mont_imgs   MM*Y x NN*X x M     big image with everything montaged
    # def montage(imgs, PAD=5, RATIO=16/9., MM=-1, NN=-1, primeDir=0,
    # verbose=False, forceFloat=False):
    if(imgs.ndim == 3):
        toExp = True
        imgs = imgs[:, :, np.newaxis, :]
    else:
        toExp = False

    Y = imgs.shape[0]
    X = imgs.shape[1]
    M = imgs.shape[2]
    N = imgs.shape[3]

    PADS = np.array((PAD))
    if(PADS.flatten().size == 1):
        PADY = PADS
        PADX = PADS
    else:
        PADY = PADS[0]
        PADX = PADS[1]

    if(MM == -1 and NN == -1):
        NN = np.ceil(np.sqrt(1.0 * N * RATIO))
        MM = np.ceil(1.0 * N / NN)
        NN = np.ceil(1.0 * N / MM)
    elif(MM == -1):
        MM = np.ceil(1.0 * N / NN)
    elif(NN == -1):
        NN = np.ceil(1.0 * N / MM)

    if(primeDir == 0):  # write top-to-bottom
        [grid_mm, grid_nn] = np.meshgrid(
            np.arange(MM, dtype='uint'), np.arange(NN, dtype='uint'))
    elif(primeDir == 1):  # write left-to-right
        [grid_nn, grid_mm] = np.meshgrid(
            np.arange(NN, dtype='uint'), np.arange(MM, dtype='uint'))

    grid_mm = np.uint(grid_mm.flatten()[0:N])
    grid_nn = np.uint(grid_nn.flatten()[0:N])

    EXTRA_PADY = EXTRA_PAD[0] * PADY
    EXTRA_PADX = EXTRA_PAD[0] * PADX

    # mont_imgs = np.zeros(((Y+PAD)*MM-PAD, (X+PAD)*NN-PAD, M), dtype=use_dtype)
    mont_imgs = np.zeros(
        (np.uint(
            (Y + PADY) * MM - PADY + EXTRA_PADY),
            np.uint(
            (X + PADX) * NN - PADX + EXTRA_PADX),
            M),
        dtype=imgs.dtype)
    mont_imgs = mont_imgs + \
        backClr.flatten()[np.newaxis, np.newaxis, :].astype(mont_imgs.dtype)

    for ii in np.random.permutation(N):
        # print imgs[:,:,:,ii].shape
        # mont_imgs[grid_mm[ii]*(Y+PAD):(grid_mm[ii]*(Y+PAD)+Y), grid_nn[ii]*(X+PAD):(grid_nn[ii]*(X+PAD)+X),:]
        mont_imgs[np.uint(grid_mm[ii] *
                          (Y +
                           PADY)):np.uint((grid_mm[ii] *
                                           (Y +
                                            PADY) +
                                           Y)), np.uint(grid_nn[ii] *
                                                        (X +
                                                         PADX)):np.uint((grid_nn[ii] *
                                                                         (X +
                                                                          PADX) +
                                                                         X)), :] = imgs[:, :, :, ii]

    if(M == 1):
        imgs = imgs.reshape(imgs.shape[0], imgs.shape[1], imgs.shape[3])

    if(toExp):
        mont_imgs = mont_imgs[:, :, 0]

    if(returnGridPos):
        # return (mont_imgs,np.concatenate((grid_mm[:,:,np.newaxis]*(Y+PAD),
        # grid_nn[:,:,np.newaxis]*(X+PAD)),axis=2))
        return (mont_imgs, np.concatenate(
            (grid_mm[:, np.newaxis] * (Y + PADY), grid_nn[:, np.newaxis] * (X + PADX)), axis=1))
        # return (mont_imgs, (grid_mm,grid_nn))
    else:
        return mont_imgs

class zeroClipper(object):
    def __init__(self, frequency=1):
        self.frequency = frequency

    def __call__(self, module):
        embed()
        if hasattr(module, 'weight'):
            # module.weight.data = torch.max(module.weight.data, 0)
            module.weight.data = torch.max(module.weight.data, 0) + 100

def flatten_nested_list(nested_list):
    # only works for list of list
    accum = []
    for sublist in nested_list:
        for item in sublist:
            accum.append(item)
    return accum

def read_file(in_path,list_lines=False):
    agg_str = ''
    f = open(in_path,'r')
    cur_line = f.readline()
    while(cur_line!=''):
        agg_str+=cur_line
        cur_line = f.readline()
    f.close()
    if(list_lines==False):
        return agg_str.replace('\n','')
    else:
        line_list = agg_str.split('\n')
        ret_list = []
        for item in line_list:
            if(item!=''):
                ret_list.append(item)
        return ret_list

def read_csv_file_as_text(in_path):
    agg_str = []
    f = open(in_path,'r')
    cur_line = f.readline()
    while(cur_line!=''):
        agg_str.append(cur_line)
        cur_line = f.readline()
    f.close()
    return agg_str

def random_swap(obj0,obj1):
    if(np.random.rand() < .5):
        return (obj0,obj1,0)
    else:
        return (obj1,obj0,1)

def voc_ap(rec, prec, use_07_metric=False):
    """ ap = voc_ap(rec, prec, [use_07_metric])
    Compute VOC AP given precision and recall.
    If use_07_metric is true, uses the
    VOC 07 11 point method (default:False).
    """
    if use_07_metric:
        # 11 point metric
        ap = 0.
        for t in np.arange(0., 1.1, 0.1):
            if np.sum(rec >= t) == 0:
                p = 0
            else:
                p = np.max(prec[rec >= t])
            ap = ap + p / 11.
    else:
        # correct AP calculation
        # first append sentinel values at the end
        mrec = np.concatenate(([0.], rec, [1.]))
        mpre = np.concatenate(([0.], prec, [0.]))

        # compute the precision envelope
        for i in range(mpre.size - 1, 0, -1):
            mpre[i - 1] = np.maximum(mpre[i - 1], mpre[i])

        # to calculate area under PR curve, look for points
        # where X axis (recall) changes value
        i = np.where(mrec[1:] != mrec[:-1])[0]

        # and sum (\Delta recall) * prec
        ap = np.sum((mrec[i + 1] - mrec[i]) * mpre[i + 1])
    return ap



import csv

def writeCSV(filename, lines):
    with open(filename, "w") as f:
        csvwriter = csv.writer(f)
        csvwriter.writerows(lines)

def readCSV(filename):
    lines = []
    with open(filename, "r") as f:
        csvreader = csv.reader(f)
        for line in csvreader:
            lines.append(line)
    return lines

def tryFloat(value):
    try:
        value = float(value)
    except:
        value = value
    
    return value

def getColumn(lines, columnid, elementType=''):
    column = []
    for line in lines:
        try:
            value = line[columnid]
        except:
            continue
            
        if elementType == 'float':
            value = tryFloat(value)

        column.append(value)
    return column



import os
import nrrd
import h5py 
import numpy as np
import torch
import logging
from collections import OrderedDict
from utils.util import calculate_psnr, calculate_ssim, calculate_pdist
from utils.util import setup_logger, print_metrics
from utils.util import create_pdist_model


def clip_to_range(sample, raw_data_range=1500):
    img = np.clip(np.array(sample, np.float32, copy=False), 0, raw_data_range)
    return img.astype(np.int16)

def read_file(file_path, ext):
    if ext == '.h5':
        with h5py.File(file_path, 'r') as file:
            data = file['data'][()]
    elif ext == '.nrrd':
        data, _ = nrrd.read(file_path)
        data = data.transpose((2, 1, 0))
    else:
        raise ValueError('Input file extension not supported!')
    return data


def main():
    if UID_file:
        with open(UID_file, 'r') as f:
            lines = f.readlines()
            uids = [l.rstrip()+LR_datatype for l in lines]
    else:
        uids = [file for file in os.listdir(path_to_LR) if file.endswith(LR_datatype)]
    # dict for metrics
    pnsr_results = OrderedDict()
    ssim_results = OrderedDict()
    pdist_results = OrderedDict()
    # create pdist model vgg
    pdist_model = create_pdist_model(use_gpu=True)
    # iterate for each cases
    for uid in uids:
        patient_id = uid.split(LR_datatype)[0]
        pnsr_results[patient_id] = {}
        ssim_results[patient_id] = {}
        pdist_results[patient_id] = {}

        # set-up data paths
        LR_case_path = os.path.join(path_to_LR, uid)
        if split_uid_for_ref:
            HR_case_path = os.path.join(path_to_HR, uid.split('_')[0]+HR_datatype)
        else:
            HR_case_path = os.path.join(path_to_HR, uid.split(LR_datatype)[0]+HR_datatype)

        # read LR-HR pair
        LR_data = read_file(LR_case_path, LR_datatype)
        HR_data = read_file(HR_case_path, HR_datatype)
        assert LR_data.shape == HR_data.shape

        sr_vol = clip_to_range(LR_data, raw_data_range=1500.)
        gt_vol = clip_to_range(HR_data, raw_data_range=1500.)
        min_depth = min(sr_vol.shape[0], gt_vol.shape[0])
        sr_vol = sr_vol[:min_depth,...]
        gt_vol = gt_vol[:min_depth,...] # make sure they have the same depth

        # compute metric
        def _calculate_metrics(sr_vol, gt_vol, view='xy'):
            sum_psnr = 0.
            sum_ssim = 0.
            sum_pdist = 0.
            # [D,H,W]
            num_val = 0 # psnr could be inf at xz or yz (near edges), will not calculate
            for i, vol in enumerate(zip(sr_vol, gt_vol)):
                sr_img, gt_img = vol[0], vol[1] # shape: [823, 512, 512] of type numpy.ndarray
                # print('i is {}, sr_img shape is {} and gt_img shape is {}'.format(i, sr_img.shape, gt_img.shape))
                # range is assume to be [0,255] so  have to scale back from 1500 to 255 float64
                crop_size = round(1.0)
                # shape [510, 510]
                cropped_sr_img = sr_img[crop_size:-crop_size, crop_size:-crop_size]\
                                    .astype(np.float64) / 1500. * 255.
                cropped_gt_img = gt_img[crop_size:-crop_size, crop_size:-crop_size]\
                                    .astype(np.float64) / 1500. * 255.
                psnr = calculate_psnr(cropped_sr_img, cropped_gt_img)
                ssim = calculate_ssim(cropped_sr_img, cropped_gt_img)
                pdist = calculate_pdist(pdist_model, cropped_sr_img, cropped_gt_img)
                if psnr != float('inf'):
                    num_val += 1
                    sum_psnr += psnr
                    sum_ssim += ssim
                    sum_pdist += pdist
                logger.info('{:20s} - {:3d}- PSNR: {:.6f} dB; SSIM: {:.6f}; pdist: {:.6f}'\
                                .format(patient_id, i+1, psnr, ssim, pdist))

            pnsr_results[patient_id][view] = sum_psnr / num_val
            ssim_results[patient_id][view] = sum_ssim / num_val
            pdist_results[patient_id][view] = sum_pdist / num_val
            return pnsr_results, ssim_results, pdist_results

        # [H W] axial view
        _calculate_metrics(sr_vol, gt_vol, view='xy')
        # [D W] coronal view
        _calculate_metrics(sr_vol.transpose(1, 0, 2), gt_vol.transpose(1, 0, 2), view='xz')
        # [D H] sagittal view
        _calculate_metrics(sr_vol.transpose(2, 0, 1), gt_vol.transpose(2, 0, 1), view='yz')

    print_metrics(logger, 'test PSNR', pnsr_results)
    print_metrics(logger, 'test SSIM', ssim_results)
    print_metrics(logger, 'test pdist', pdist_results)    
    logger.info('All metrics computed!')



if __name__ == '__main__':
    # set-up data types
    path_to_LR = '/datasets/data_st1.0_merged'
    LR_datatype = '.h5'
    UID_file = '/datasets/uids/70-30_split/ucla_testUID_70_30_split.txt' # if UID provided, read only specific cases

    path_to_HR = '/datasets/reference/k2_d100_st1'
    HR_datatype = '.h5'
    split_uid_for_ref = True

    # set-up logger
    path_to_save_metric = '/workspace/cNormGAN-AC-LT/results/test_Unnorm-70-30-Split_tileStitch-32x64x64_Ref-k2d100/merged_cases'
    setup_logger(None, path_to_save_metric, 'test.log', level=logging.INFO, screen=True)
    logger = logging.getLogger('base')

    main()





import sys

from .layer import *

from lung_analysis_pipeline.NoduleDetect_SANet.config import net_config as config
from lung_analysis_pipeline.utils.util import rescale_image

import copy
from torch.nn.parallel.data_parallel import data_parallel
import time
import torch.nn.functional as F
from lung_analysis_pipeline.NoduleDetect_SANet.utils.util import center_box_to_coord_box, ext2factor, clip_boxes, pad2factor
from torch.nn.parallel import data_parallel
import random
from scipy.stats import norm
from . import resnet, cgnl

bn_momentum = 0.1
affine = True


class ResBlock3d(nn.Module):
    def __init__(self, n_in, n_out, stride = 1):
        super(ResBlock3d, self).__init__()
        self.conv1 = nn.Conv3d(n_in, n_out, kernel_size = 3, stride = stride, padding = 1)
        self.bn1 = nn.BatchNorm3d(n_out, momentum=bn_momentum)
        self.relu = nn.ReLU(inplace = True)
        self.conv2 = nn.Conv3d(n_out, n_out, kernel_size = 3, padding = 1)
        self.bn2 = nn.BatchNorm3d(n_out, momentum=bn_momentum)

        if stride != 1 or n_out != n_in:
            self.shortcut = nn.Sequential(
                nn.Conv3d(n_in, n_out, kernel_size = 1, stride = stride),
                nn.BatchNorm3d(n_out, momentum=bn_momentum))
        else:
            self.shortcut = None

    def forward(self, x):
        residual = x
        if self.shortcut is not None:
            residual = self.shortcut(x)
        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)
        out = self.conv2(out)
        out = self.bn2(out)

        out += residual
        out = self.relu(out)
        return out


class FeatureNet(nn.Module):
    def __init__(self, config):
        super(FeatureNet, self).__init__()
        self.resnet50 = resnet.resnet50()

        self.back1 = nn.Sequential(
            nn.Conv3d(256, 128, kernel_size=3, padding=1, stride=1),
            nn.BatchNorm3d(128, momentum=bn_momentum),
            nn.ReLU(inplace=True))

        self.back2 = nn.Sequential(
            nn.Conv3d(128, 128, kernel_size=3, padding=1, stride=1),
            nn.BatchNorm3d(128, momentum=bn_momentum),
            nn.ReLU(inplace=True))

        self.back3 = nn.Sequential(
            nn.Conv3d(128, 64, kernel_size=3, padding=1, stride=1),
            nn.BatchNorm3d(64, momentum=bn_momentum),
            nn.ReLU(inplace=True))

        # upsampling
        self.reduce1 = nn.Sequential(
            nn.Conv3d(64, 64, kernel_size=1, stride=1),
            nn.BatchNorm3d(64, momentum=bn_momentum),
            nn.ReLU(inplace=True))

        self.path1 = nn.Sequential(
            nn.ConvTranspose3d(64, 64, kernel_size=2, stride=2),
            nn.BatchNorm3d(64),
            nn.ReLU(inplace=True))

        self.reduce2 = nn.Sequential(
            nn.Conv3d(64, 64, kernel_size=1, stride=1),
            nn.BatchNorm3d(64, momentum=bn_momentum),
            nn.ReLU(inplace=True))

        # upsampling
        self.path2 = nn.Sequential(
            nn.ConvTranspose3d(64, 64, kernel_size=2, stride=2),
            nn.BatchNorm3d(64),
            nn.ReLU(inplace=True))

        self.reduce3 = nn.Sequential(
            nn.Conv3d(64, 64, kernel_size=1, stride=1),
            nn.BatchNorm3d(64, momentum=bn_momentum),
            nn.ReLU(inplace=True))

        self.path3 = nn.Sequential(
            nn.ConvTranspose3d(128, 128, kernel_size=2, stride=2),
            nn.BatchNorm3d(128),
            nn.ReLU(inplace=True))

        self.reduce4 = nn.Sequential(
            nn.Conv3d(256, 128, kernel_size=1, stride=1),
            nn.BatchNorm3d(128, momentum=bn_momentum),
            nn.ReLU(inplace=True))

    def forward(self, x):
        x1, out1, out2, out3, out4 = self.resnet50(x)

        out4 = self.reduce1(out4)
        rev3 = self.path1(out4)
        out3 = self.reduce2(out3)
        comb3 = self.back3(torch.cat((rev3, out3), 1))
        rev2 = self.path2(comb3)
        out2 = self.reduce3(out2)
        comb2 = self.back2(torch.cat((rev2, out2), 1))

        return [x1, out1, comb2], out1

class RpnHead(nn.Module):
    def __init__(self, config, in_channels=128):
        super(RpnHead, self).__init__()
        self.drop = nn.Dropout3d(p=0.5, inplace=False)
        self.conv = nn.Sequential(nn.Conv3d(in_channels, 64, kernel_size=1),
                                    nn.ReLU())
        self.logits = nn.Conv3d(64, 1 * len(config['anchors']), kernel_size=1)
        self.deltas = nn.Conv3d(64, 6 * len(config['anchors']), kernel_size=1)

    def forward(self, f):
        # out = self.drop(f)
        out = self.conv(f)

        logits = self.logits(out)
        deltas = self.deltas(out)
        size = logits.size()
        logits = logits.view(logits.size(0), logits.size(1), -1)
        logits = logits.transpose(1, 2).contiguous().view(size[0], size[2], size[3], size[4], len(config['anchors']), 1)
        
        size = deltas.size()
        deltas = deltas.view(deltas.size(0), deltas.size(1), -1)
        deltas = deltas.transpose(1, 2).contiguous().view(size[0], size[2], size[3], size[4], len(config['anchors']), 6)
        

        return logits, deltas

class RcnnHead(nn.Module):
    def __init__(self, cfg, in_channels=128):
        super(RcnnHead, self).__init__()
        self.num_class = cfg['num_class']
        self.crop_size = cfg['rcnn_crop_size']

        self.fc1 = nn.Linear(in_channels * self.crop_size[0] * self.crop_size[1] * self.crop_size[2], 512)
        self.fc2 = nn.Linear(512, 256)
        self.logit = nn.Linear(256, self.num_class)
        self.delta = nn.Linear(256, self.num_class * 6)

    def forward(self, crops):
        x = crops.view(crops.size(0), -1)
        x = F.relu(self.fc1(x), inplace=True)
        x = F.relu(self.fc2(x), inplace=True)
        # x = F.dropout(x, 0.5, training=self.training)
        logits = self.logit(x)
        deltas = self.delta(x)

        return logits, deltas


class MaskHead(nn.Module):
    def __init__(self, cfg, in_channels=128):
        super(MaskHead, self).__init__()
        self.num_class = cfg['num_class']

        self.up1 = nn.Sequential(
            nn.Upsample(scale_factor=2, mode='trilinear'),
            nn.Conv3d(in_channels, 64, kernel_size=3, padding=1),
            nn.InstanceNorm3d(64, momentum=bn_momentum, affine=affine),
            nn.ReLU(inplace = True))
        
        self.up2 = nn.Sequential(
            nn.Upsample(scale_factor=2, mode='trilinear'),
            nn.Conv3d(in_channels, 64, kernel_size=3, padding=1),
            nn.InstanceNorm3d(64, momentum=bn_momentum, affine=affine),
            nn.ReLU(inplace = True))

        self.up3 = nn.Sequential(
            nn.Upsample(scale_factor=2, mode='trilinear'),
            nn.Conv3d(64, 64, kernel_size=3, padding=1),
            nn.InstanceNorm3d(64, momentum=bn_momentum, affine=affine),
            nn.ReLU(inplace = True))
        
        self.back1 = nn.Sequential(
            nn.Conv3d(128, 64, kernel_size=3, padding=1),
            nn.InstanceNorm3d(64, momentum=bn_momentum, affine=affine),
            nn.ReLU(inplace = True))
        
        self.back2 = nn.Sequential(
            nn.Conv3d(96, 64, kernel_size=3, padding=1),
            nn.InstanceNorm3d(64, momentum=bn_momentum, affine=affine),
            nn.ReLU(inplace = True))
        
        self.back3 = nn.Sequential(
            nn.Conv3d(65, 64, kernel_size=3, padding=1),
            nn.InstanceNorm3d(64, momentum=bn_momentum, affine=affine),
            nn.ReLU(inplace = True))

        for i in range(self.num_class):
            setattr(self, 'logits' + str(i + 1), nn.Conv3d(64, 1, kernel_size=1))

    def forward(self, detections, features):
        img, f_2, f_4 = features  

        # Squeeze the first dimension to recover from protection on avoiding split by dataparallel      
        img = img.squeeze(0)
        f_2 = f_2.squeeze(0)
        f_4 = f_4.squeeze(0)

        _, _, D, H, W = img.shape
        out = []

        for detection in detections:
            b, z_start, y_start, x_start, z_end, y_end, x_end, cat = detection

            up1 = f_4[b, :, z_start / 4:z_end / 4, y_start / 4:y_end / 4, x_start / 4:x_end / 4].unsqueeze(0)
            up2 = self.up2(up1)
            up2 = self.back2(torch.cat((up2, f_2[b, :, z_start / 2:z_end / 2, y_start / 2:y_end / 2, x_start / 2:x_end / 2].unsqueeze(0)), 1))
            up3 = self.up3(up2)
            im = img[b, :, z_start:z_end, y_start:y_end, x_start:x_end].unsqueeze(0)
            up3 = self.back3(torch.cat((up3, im), 1))

            logits = getattr(self, 'logits' + str(int(cat)))(up3)
            logits = logits.squeeze()
 
            mask = Variable(torch.zeros((D, H, W))).cuda()
            mask[z_start:z_end, y_start:y_end, x_start:x_end] = logits
            mask = mask.unsqueeze(0)
            out.append(mask)

        out = torch.cat(out, 0)

        return out


def crop_mask_regions(masks, crop_boxes):
    out = []
    for i in range(len(crop_boxes)):
        b, z_start, y_start, x_start, z_end, y_end, x_end, cat = crop_boxes[i]
        m = masks[i][z_start:z_end, y_start:y_end, x_start:x_end].contiguous()
        out.append(m)
    
    return out


def top1pred(boxes):
    res = []
    pred_cats = np.unique(boxes[:, -1])
    for cat in pred_cats:
        preds = boxes[boxes[:, -1] == cat]
        res.append(preds[0])
        
    res = np.array(res)
    return res

def random1pred(boxes):
    res = []
    pred_cats = np.unique(boxes[:, -1])
    for cat in pred_cats:
        preds = boxes[boxes[:, -1] == cat]
        idx = random.sample(range(len(preds)), 1)[0]
        res.append(preds[idx])
        
    res = np.array(res)
    return res


class CropRoi(nn.Module):
    def __init__(self, cfg, rcnn_crop_size, in_channels = 128):
        super(CropRoi, self).__init__()
        self.cfg = cfg
        self.rcnn_crop_size  = rcnn_crop_size
        self.scale = cfg['stride']
        self.DEPTH, self.HEIGHT, self.WIDTH = cfg['crop_size']

        self.up2 = nn.Sequential(
            nn.Upsample(scale_factor=2, mode='trilinear', align_corners=True),
            nn.Conv3d(in_channels, 64, kernel_size=1, padding=0),
            nn.InstanceNorm3d(64, momentum=bn_momentum, affine=affine),
            nn.ReLU(inplace=True))
        self.up3 = nn.Sequential(
            nn.Upsample(scale_factor=2, mode='trilinear',align_corners=True),
            nn.Conv3d(64, 64, kernel_size=1, padding=0),
            nn.InstanceNorm3d(64, momentum=bn_momentum, affine=affine),
            nn.ReLU(inplace=True))
        self.back2 = nn.Sequential(
            nn.Conv3d(128, 64, kernel_size=1, padding=0),
            nn.InstanceNorm3d(64, momentum=bn_momentum, affine=affine),
            nn.ReLU(inplace=True))
        self.back3 = nn.Sequential(
            nn.Conv3d(65, 64, kernel_size=3, padding=1),
            nn.InstanceNorm3d(64, momentum=bn_momentum, affine=affine),
            nn.ReLU(inplace=True))

    def forward(self, f, inputs, proposals):
        img, out1, comb2 = f
        self.DEPTH, self.HEIGHT, self.WIDTH = inputs.shape[2:]

        img = img.squeeze(0)
        out1 = out1.squeeze(0)
        comb2 = comb2.squeeze(0)

        crops = []
        for p in proposals:

            b, z_start, y_start, x_start, z_end, y_end, x_end = p

            # Slice 0 dim, should never happen
            c0 = np.array([z_start, y_start, x_start])
            c1 = np.array([z_end, y_end, x_end])
            if np.any((c1 - c0) < 1): #np.any((c1 - c0).cpu().data.numpy() < 1):
                # c0=c0+1
                # c1=c1+1
                for i in range(3):
                    if c1[i] == 0:
                        c1[i] = c1[i] + 4
                    if c1[i] - c0[i] == 0:
                        c1[i] = c1[i] + 4
                # print(p)
                # print('c0:', c0, ', c1:', c1)
            z_end, y_end, x_end = c1

            fe1 = comb2[b, :, (z_start / 4).long():(z_end / 4).long(), (y_start / 4).long():(y_end / 4).long(), (x_start / 4).long():(x_end / 4).long()].unsqueeze(0)
            fe1_up = self.up2(fe1)

            fe2 = self.back2(torch.cat((fe1_up, out1[b, :, (z_start / 2).long():(z_end / 2).long(), (y_start / 2).long():(y_end / 2).long(), (x_start / 2).long():(x_end / 2).long()].unsqueeze(0)), 1))
            # fe2_up = self.up3(fe2)

            # im = img[b, :, z_start / 2:z_end / 2, y_start / 2:y_end / 2, x_start / 2:x_end / 2].unsqueeze(0)
            # up3 = self.back3(torch.cat((fe2, im), 1))
            # crop = up3.squeeze()

            crop = fe2.squeeze()
            # crop = f[b, :, c0[0]:c1[0], c0[1]:c1[1], c0[2]:c1[2]]
            crop = F.adaptive_max_pool3d(crop, self.rcnn_crop_size)
            crops.append(crop)

        crops = torch.stack(crops)

        return crops

class SANet(nn.Module):
    def __init__(self, cfg, mode='eval'):
        super(SANet, self).__init__()

        self.cfg = cfg
        self.mode = mode
        self.feature_net = FeatureNet(config)
        self.rpn = RpnHead(config, in_channels=128)
        self.rcnn_head = RcnnHead(config, in_channels=64)
        self.rcnn_crop = CropRoi(self.cfg, cfg['rcnn_crop_size'])
        self.use_rcnn = False
        # self.rpn_loss = Loss(cfg['num_hard'])
        

    def forward(self, inputs, truth_boxes, truth_labels, split_combiner=None, nzhw=None):
        
        # features, feat_4 = self.feature_net(inputs)
        features, feat_4 = data_parallel(self.feature_net, (inputs))
        fs = features[-1]
        self.rpn_logits_flat, self.rpn_deltas_flat = data_parallel(self.rpn, fs)

        b,D,H,W,_,num_class = self.rpn_logits_flat.shape

        self.rpn_logits_flat = self.rpn_logits_flat.view(b, -1, 1)
        self.rpn_deltas_flat = self.rpn_deltas_flat.view(b, -1, 6)
        self.rpn_window = make_rpn_windows(fs, self.cfg)

        self.rpn_proposals = []
        if self.use_rcnn or self.mode in ['eval', 'test']:
            self.rpn_proposals = rpn_nms(self.cfg, self.mode, inputs, self.rpn_window,
                  self.rpn_logits_flat, self.rpn_deltas_flat)
        if self.mode in ['train', 'valid']:
            # self.rpn_proposals = torch.zeros((0, 8)).cuda()
            self.rpn_labels, self.rpn_label_assigns, self.rpn_label_weights, self.rpn_targets, self.rpn_target_weights = \
                make_rpn_target(self.cfg, self.mode, inputs, self.rpn_window, truth_boxes, truth_labels )

            if self.use_rcnn:
                # self.rpn_proposals = torch.zeros((0, 8)).cuda()
                self.rpn_proposals, self.rcnn_labels, self.rcnn_assigns, self.rcnn_targets = \
                    make_rcnn_target(self.cfg, self.mode, inputs, self.rpn_proposals,
                        truth_boxes, truth_labels)

        #rcnn proposals
        self.detections = copy.deepcopy(self.rpn_proposals)
        self.ensemble_proposals = copy.deepcopy(self.rpn_proposals)

        self.mask_probs = []
        if self.use_rcnn:
            if len(self.rpn_proposals) > 0:
                proposal = self.rpn_proposals[:, [0, 2, 3, 4, 5, 6, 7]].cpu().numpy().copy()
                proposal[:, 1:] = center_box_to_coord_box(proposal[:, 1:])
                proposal = proposal.astype(np.int64)
                proposal[:, 1:] = ext2factor(proposal[:, 1:], 4)
                proposal[:, 1:] = clip_boxes(proposal[:, 1:], inputs.shape[2:])
                # rcnn_crops = self.rcnn_crop(features, inputs, torch.from_numpy(proposal).cuda())
                features = [t.unsqueeze(0).expand(torch.cuda.device_count(), -1, -1, -1, -1, -1) for t in features]
                rcnn_crops = data_parallel(self.rcnn_crop, (features, inputs, torch.from_numpy(proposal).cuda()))
                # rcnn_crops = self.rcnn_crop(fs, inputs, self.rpn_proposals)
                # self.rcnn_logits, self.rcnn_deltas = self.rcnn_head(rcnn_crops)
                self.rcnn_logits, self.rcnn_deltas = data_parallel(self.rcnn_head, rcnn_crops)
                self.detections, self.keeps = rcnn_nms(self.cfg, self.mode, inputs, self.rpn_proposals, 
                                                                        self.rcnn_logits, self.rcnn_deltas)

                if self.mode in ['eval']:
                    #     Ensemble
                    fpr_res = get_probability(self.cfg, self.mode, inputs, self.rpn_proposals, self.rcnn_logits,
                                              self.rcnn_deltas)
                    if self.ensemble_proposals.shape[0] == fpr_res.shape[0]:
                        self.ensemble_proposals[:, 1] = (self.ensemble_proposals[:, 1] + fpr_res[:, 0]) / 2
    
    # for pipeline 
    def run_test(self, input_data): 
        self.set_mode('eval')
        self.use_rcnn = True 

        # preprocess the volume 
        inputs = input_data['volume']
        inputs = inputs.cuda()

        inputs = pad2factor(inputs[0,0].cpu().numpy()) # padding is required for SANet 
        inputs = rescale_image(inputs, min_value=-1, max_value=1) # input image might be (0,1500) or (0,1). Need to rescale to (-1,1)
        inputs = torch.from_numpy(inputs)[None,None].cuda()
        
        # try: 
        with torch.no_grad(): 
            self.forward(inputs, None, None)
        detections = self.rpn_proposals.cpu().numpy() 
        if len(detections) == 0:
            detections = np.empty((0, 0), dtype=np.float32)
        else: 
            detections = detections[:, 1:-1]
        
        input_data['detection_bboxes'] = torch.from_numpy(detections)[None] 
        return input_data 
        # except Exception as e: 
        #     print("An exception occurred:", str(e))
        #     torch.cuda.empty_cache() 

    def loss(self, targets=None):
        cfg  = self.cfg
    
        self.rcnn_cls_loss, self.rcnn_reg_loss = torch.zeros(1).cuda(), torch.zeros(1).cuda()
        rcnn_stats = None
    
        self.rpn_cls_loss, self.rpn_reg_loss, rpn_stats = \
           rpn_loss( self.rpn_logits_flat, self.rpn_deltas_flat, self.rpn_labels,
            self.rpn_label_weights, self.rpn_targets, self.rpn_target_weights, self.cfg, mode=self.mode)
    
        if self.use_rcnn:
            self.rcnn_cls_loss, self.rcnn_reg_loss, rcnn_stats = \
                rcnn_loss(self.rcnn_logits, self.rcnn_deltas, self.rcnn_labels, self.rcnn_targets)
    
        self.total_loss = self.rpn_cls_loss + self.rpn_reg_loss \
                          + self.rcnn_cls_loss +  self.rcnn_reg_loss

    
        return self.total_loss, rpn_stats, rcnn_stats

    def set_mode(self, mode):
        assert mode in ['train', 'valid', 'eval', 'test']
        self.mode = mode
        if mode in ['train']:
            self.train()
        else:
            self.eval()

    def set_anchor_params(self, anchor_ids, anchor_params):
        self.anchor_ids = anchor_ids
        self.anchor_params = anchor_params

    def crf(self, detections):
        """
        detections: numpy array of detection results [b, z, y, x, d, h, w, p]
        """
        res = []
        config = self.cfg
        anchor_ids = self.anchor_ids
        anchor_params = self.anchor_params
        anchor_centers = []

        for a in anchor_ids:
            # category starts from 1 with 0 denoting background
            # id starts from 0
            cat = a + 1
            dets = detections[detections[:, -1] == cat]
            if len(dets):
                b, p, z, y, x, d, h, w, _ = dets[0]
                anchor_centers.append([z, y, x])
                res.append(dets[0])
            else:
                # Does not have anchor box
                return detections
        
        pred_cats = np.unique(detections[:, -1]).astype(np.uint8)
        for cat in pred_cats:
            if cat - 1 not in anchor_ids:
                cat = int(cat)
                preds = detections[detections[:, -1] == cat]
                score = np.zeros((len(preds),))
                roi_name = config['roi_names'][cat - 1]

                for k, params in enumerate(anchor_params):
                    param = params[roi_name]
                    for i, det in enumerate(preds):
                        b, p, z, y, x, d, h, w, _ = det
                        d = np.array([z, y, x]) - np.array(anchor_centers[k])
                        prob = norm.pdf(d, param[0], param[1])
                        prob = np.log(prob)
                        prob = np.sum(prob)
                        score[i] += prob

                res.append(preds[score == score.max()][0])
            
        res = np.array(res)
        return res

# if __name__ == '__main__':
#     net = SANet(config)

#     input = torch.rand([2,1,304,384,384])
#     input = Variable(input)
#     net(input, None, None)




from net.lib.box.nms.torch_nms.extension import *

import torch
import numpy as np

def torch_nms(dets, thresh):
    """
    dets has to be a tensor
    """
    if not dets.is_cuda:
        z = dets[:, 1]
        y = dets[:, 2]
        x = dets[:, 3]
        d = dets[:, 4]
        h = dets[:, 5]
        w = dets[:, 6]
        scores = dets[:, 0]

        areas = d * h * w
        order = scores.sort(0, descending=True)[1]
        # order = torch.from_numpy(np.ascontiguousarray(scores.numpy().argsort()[::-1])).long()

        keep = torch.LongTensor(dets.size(0))
        num_out = torch.LongTensor(1)
        cpu_nms(keep, num_out, dets, order, areas, thresh)

        return dets[keep[:num_out[0]]], keep[:num_out[0]]

    else:
        raise NotImplementedError
        z = dets[:, 1]
        y = dets[:, 2]
        x = dets[:, 3]
        d = dets[:, 4]
        scores = dets[:, 0]

        areas = d ** 3
        order = scores.sort(0, descending=True)[1]
        # order = torch.from_numpy(np.ascontiguousarray(scores.cpu().numpy().argsort()[::-1])).long().cuda()

        dets = dets[order].contiguous()

        keep = torch.LongTensor(dets.size(0))
        num_out = torch.LongTensor(1)
        # keep = torch.cuda.LongTensor(dets.size(0))
        # num_out = torch.cuda.LongTensor(1)
        gpu_nms(keep, num_out, dets, thresh)

        # return dets[keep[:num_out[0]].cuda()].contiguous(), keep[:num_out[0]].cuda()
        return dets[keep[:num_out[0]].cuda()].contiguous(), order[keep[:num_out[0]]].contiguous()




import numpy as np
import os
import time
from . import util
from . import html
# from pdb import set_trace as st
import matplotlib.pyplot as plt
import math
# from IPython import embed

def zoom_to_res(img,res=256,order=0,axis=0):
    # img   3xXxX
    from scipy.ndimage import zoom
    zoom_factor = res/img.shape[1]
    if(axis==0):
        return zoom(img,[1,zoom_factor,zoom_factor],order=order)
    elif(axis==2):
        return zoom(img,[zoom_factor,zoom_factor,1],order=order)

class Visualizer():
    def __init__(self, opt):
        # self.opt = opt
        self.display_id = opt.display_id
        # self.use_html = opt.is_train and not opt.no_html
        self.win_size = opt.display_winsize
        self.name = opt.name
        self.display_cnt = 0 # display_current_results counter
        self.display_cnt_high = 0
        self.use_html = opt.use_html

        if self.display_id > 0:
            import visdom
            self.vis = visdom.Visdom(port = opt.display_port)

        self.web_dir = os.path.join(opt.checkpoints_dir, opt.name, 'web')
        util.mkdirs([self.web_dir,])
        if self.use_html:
            self.img_dir = os.path.join(self.web_dir, 'images')
            print('create web directory %s...' % self.web_dir)
            util.mkdirs([self.img_dir,])

    # |visuals|: dictionary of images to display or save
    def display_current_results(self, visuals, epoch, nrows=None, res=256):
        if self.display_id > 0: # show images in the browser
            title = self.name
            if(nrows is None):
                nrows = int(math.ceil(len(visuals.items()) / 2.0))
            images = []
            idx = 0
            for label, image_numpy in visuals.items():
                title += " | " if idx % nrows == 0 else ", "
                title += label
                img = image_numpy.transpose([2, 0, 1])
                img = zoom_to_res(img,res=res,order=0)
                images.append(img)
                idx += 1
            if len(visuals.items()) % 2 != 0:
                white_image = np.ones_like(image_numpy.transpose([2, 0, 1]))*255
                white_image = zoom_to_res(white_image,res=res,order=0)
                images.append(white_image)
            self.vis.images(images, nrow=nrows, win=self.display_id + 1,
                            opts=dict(title=title))

        if self.use_html: # save images to a html file
            for label, image_numpy in visuals.items():
                img_path = os.path.join(self.img_dir, 'epoch%.3d_cnt%.6d_%s.png' % (epoch, self.display_cnt, label))
                util.save_image(zoom_to_res(image_numpy, res=res, axis=2), img_path)

            self.display_cnt += 1
            self.display_cnt_high = np.maximum(self.display_cnt_high, self.display_cnt)

            # update website
            webpage = html.HTML(self.web_dir, 'Experiment name = %s' % self.name, reflesh=1)
            for n in range(epoch, 0, -1):
                webpage.add_header('epoch [%d]' % n)
                if(n==epoch):
                    high = self.display_cnt
                else:
                    high = self.display_cnt_high
                for c in range(high-1,-1,-1):
                    ims = []
                    txts = []
                    links = []

                    for label, image_numpy in visuals.items():
                        img_path = 'epoch%.3d_cnt%.6d_%s.png' % (n, c, label)
                        ims.append(os.path.join('images',img_path))
                        txts.append(label)
                        links.append(os.path.join('images',img_path))
                    webpage.add_images(ims, txts, links, width=self.win_size)
            webpage.save()

    # save errors into a directory
    def plot_current_errors_save(self, epoch, counter_ratio, opt, errors,keys='+ALL',name='loss', to_plot=False):
        if not hasattr(self, 'plot_data'):
            self.plot_data = {'X':[],'Y':[], 'legend':list(errors.keys())}
        self.plot_data['X'].append(epoch + counter_ratio)
        self.plot_data['Y'].append([errors[k] for k in self.plot_data['legend']])

        # embed()
        if(keys=='+ALL'):
            plot_keys = self.plot_data['legend']
        else:
            plot_keys = keys

        if(to_plot):
            (f,ax) = plt.subplots(1,1)
        for (k,kname) in enumerate(plot_keys):
            kk = np.where(np.array(self.plot_data['legend'])==kname)[0][0]
            x = self.plot_data['X']
            y = np.array(self.plot_data['Y'])[:,kk]
            if(to_plot):
                ax.plot(x, y, 'o-', label=kname)
            np.save(os.path.join(self.web_dir,'%s_x')%kname,x)
            np.save(os.path.join(self.web_dir,'%s_y')%kname,y)

        if(to_plot):
            plt.legend(loc=0,fontsize='small')
            plt.xlabel('epoch')
            plt.ylabel('Value')
            f.savefig(os.path.join(self.web_dir,'%s.png'%name))
            f.clf()
            plt.close()

    # errors: dictionary of error labels and values
    def plot_current_errors(self, epoch, counter_ratio, opt, errors):
        if not hasattr(self, 'plot_data'):
            self.plot_data = {'X':[],'Y':[], 'legend':list(errors.keys())}
        self.plot_data['X'].append(epoch + counter_ratio)
        self.plot_data['Y'].append([errors[k] for k in self.plot_data['legend']])
        self.vis.line(
            X=np.stack([np.array(self.plot_data['X'])]*len(self.plot_data['legend']),1),
            Y=np.array(self.plot_data['Y']),
            opts={
                'title': self.name + ' loss over time',
                'legend': self.plot_data['legend'],
                'xlabel': 'epoch',
                'ylabel': 'loss'},
            win=self.display_id)

    # errors: same format as |errors| of plotCurrentErrors
    def print_current_errors(self, epoch, i, errors, t, t2=-1, t2o=-1, fid=None):
        message = '(ep: %d, it: %d, t: %.3f[s], ept: %.2f/%.2f[h]) ' % (epoch, i, t, t2o, t2)
        message += (', ').join(['%s: %.3f' % (k, v) for k, v in errors.items()])

        print(message)
        if(fid is not None):
            fid.write('%s\n'%message)


    # save image to the disk
    def save_images_simple(self, webpage, images, names, in_txts, prefix='', res=256):
        image_dir = webpage.get_image_dir()
        ims = []
        txts = []
        links = []

        for name, image_numpy, txt in zip(names, images, in_txts):
            image_name = '%s_%s.png' % (prefix, name)
            save_path = os.path.join(image_dir, image_name)
            if(res is not None):
                util.save_image(zoom_to_res(image_numpy,res=res,axis=2), save_path)
            else:
                util.save_image(image_numpy, save_path)

            ims.append(os.path.join(webpage.img_subdir,image_name))
            # txts.append(name)
            txts.append(txt)
            links.append(os.path.join(webpage.img_subdir,image_name))
        # embed()
        webpage.add_images(ims, txts, links, width=self.win_size)

    # save image to the disk
    def save_images(self, webpage, images, names, image_path, title=''):
        image_dir = webpage.get_image_dir()
        # short_path = ntpath.basename(image_path)
        # name = os.path.splitext(short_path)[0]
        # name = short_path
        # webpage.add_header('%s, %s' % (name, title))
        ims = []
        txts = []
        links = []

        for label, image_numpy in zip(names, images):
            image_name = '%s.jpg' % (label,)
            save_path = os.path.join(image_dir, image_name)
            util.save_image(image_numpy, save_path)

            ims.append(image_name)
            txts.append(label)
            links.append(image_name)
        webpage.add_images(ims, txts, links, width=self.win_size)

    # save image to the disk
    # def save_images(self, webpage, visuals, image_path, short=False):
    #     image_dir = webpage.get_image_dir()
    #     if short:
    #         short_path = ntpath.basename(image_path)
    #         name = os.path.splitext(short_path)[0]
    #     else:
    #         name = image_path

    #     webpage.add_header(name)
    #     ims = []
    #     txts = []
    #     links = []

    #     for label, image_numpy in visuals.items():
    #         image_name = '%s_%s.png' % (name, label)
    #         save_path = os.path.join(image_dir, image_name)
    #         util.save_image(image_numpy, save_path)

    #         ims.append(image_name)
    #         txts.append(label)
    #         links.append(image_name)
    #     webpage.add_images(ims, txts, links, width=self.win_size)



import sys
import numpy as np
import torch
from lung_analysis_pipeline.NoduleDetect_SANet.config import config
import pydicom as dicom
import numpy as np
from scipy.sparse import csc_matrix
from collections import defaultdict
import os
import shutil
import operator
import warnings
import numpy as np
# import matplotlib as mpl
# mpl.use('TkAgg')
# import matplotlib.pyplot as plt
# import SimpleITK as sitk
import pydicom
import matplotlib.cm as cm
import math
from skimage import measure
from scipy.ndimage import zoom
from scipy.sparse import csr_matrix
from scipy.spatial.distance import cdist
import pandas as pd
import cv2
try:
    # Python2
    from StringIO import StringIO
except ImportError:
    # Python3
    from io import StringIO


class Logger(object):
    def __init__(self,logfile):
        self.terminal = sys.stdout
        self.log = open(logfile, "a")

    def write(self, message):
        self.terminal.write(message)
        self.log.write(message)

    def flush(self):
        #this flush method is needed for python 3 compatibility.
        #this handles the flush command by doing nothing.
        #you might want to specify some extra behavior here.
        pass


def worldToVoxelCoord(worldCoord, origin, spacing):
    stretchedVoxelCoord = np.absolute(worldCoord - origin)
    voxelCoord = stretchedVoxelCoord / spacing
    return voxelCoord


def voxelToWorldCoord(voxelCoord, origin, spacing):
    worldCoord = voxelCoord * spacing
    worldCoord += origin
    return worldCoord


def npy2submission(set_name, save_path, bbox_dir, prep_dir, postfix='detection'):
    """
    :param set_name: lists of patient names
    :param save_path: path to save the submission file
    :param bbox_dir: directory saving predicted bounded boxes
    :param prep_dir: directory saving preprocessing results, should include *_origin.npy, *_spacing.npy,
                     *_ebox_origin.npy
    :return: None
    """
    filenames = np.genfromtxt(set_name, dtype=str)
    if not filenames.shape:
        filenames = np.array([filenames])

    k = 1000
#     name_map = pd.read_csv('luna_shorter.csv', dtype=str)
    # print name_map

    submission = []

    for name in filenames:
        pbb = np.load(os.path.join(bbox_dir, "%s_%s.npy" % (name, postfix)))
        if len(pbb) == 0:
            continue
        pbb = pbb[pbb[:, 0].argsort()][::-1][:k]
        spacing = np.load(os.path.join(prep_dir, name + '_spacing.npy'))
        ebox_origin = np.load(os.path.join(prep_dir, name + '_ebox_origin.npy'))
        origin = np.load(os.path.join(prep_dir, name + '_origin.npy'))

        for p in pbb:
            ebox_coord = p[[1, 2, 3]]
            whole_img_coord = ebox_coord + ebox_origin
            worldCoord = voxelToWorldCoord(whole_img_coord, origin, spacing)
            submission.append([name, worldCoord[2], worldCoord[1], worldCoord[0], p[0]])

    submission = pd.DataFrame(submission, columns=["seriesuid", "coordX", "coordY", "coordZ",
                                                   "probability"])

    print("Saving submission to", save_path)
    submission.to_csv(save_path, sep=',', index=False, header=True)


def center_box_to_coord_box(bboxes):
    """
    Convert bounding box using center of rectangle and side lengths representation to 
    bounding box using coordinate representation
    [center_z, center_y, center_x, D, H, W] -> [z_start, y_start, x_start, z_end, y_end, x_end]

    bboxes: list of bounding boxes, [num_bbox, 6]
    """
    res = np.zeros(bboxes.shape)
    res[:, 0] = bboxes[:, 0] - bboxes[:, 3] / 2.
    res[:, 1] = bboxes[:, 1] - bboxes[:, 4] / 2.
    res[:, 2] = bboxes[:, 2] - bboxes[:, 5] / 2.
    res[:, 3] = bboxes[:, 0] + bboxes[:, 3] / 2.
    res[:, 4] = bboxes[:, 1] + bboxes[:, 4] / 2.
    res[:, 5] = bboxes[:, 2] + bboxes[:, 5] / 2.

    return res


def coord_box_to_center_box(bboxes):
    """
    Convert bounding box using coordinate representation to 
    bounding box using center of rectangle and side lengths representation
    [z_start, y_start, x_start, z_end, y_end, x_end] -> [center_z, center_y, center_x, D, H, W]

    bboxes: list of bounding boxes, [num_bbox, 6]
    """
    res = np.zeros(bboxes.shape)

    res[:, 3] = bboxes[:, 3] - bboxes[:, 0]
    res[:, 4] = bboxes[:, 4] - bboxes[:, 1]
    res[:, 5] = bboxes[:, 5] - bboxes[:, 2]
    res[:, 0] = bboxes[:, 0] + res[:, 3] / 2.
    res[:, 1] = bboxes[:, 1] + res[:, 4] / 2.
    res[:, 2] = bboxes[:, 2] + res[:, 5] / 2.

    return res

def ext2factor(bboxes, factor=8):
    """
    Given center box representation which is [z_start, y_start, x_start, z_end, y_end, x_end],
    return closest point which can be divided by 8 
    """
    bboxes[:, :3] = bboxes[:, :3] // factor * factor
    bboxes[:, 3:] = bboxes[:, 3:] // factor * factor + (bboxes[:, 3:] % factor != 0).astype(np.int32) * factor

    return bboxes

def clip_boxes(boxes, img_size):
    '''
    clip boxes outside the image, all box follows [z_start, y_start, x_start, z_end, y_end, x_end]
    '''
    depth, height, width = img_size
    boxes[:, 0] = np.clip(boxes[:, 0], 0, depth)
    boxes[:, 1] = np.clip(boxes[:, 1], 0, height)
    boxes[:, 2] = np.clip(boxes[:, 2], 0, width)
    boxes[:, 3] = np.clip(boxes[:, 3], 0, depth)
    boxes[:, 4] = np.clip(boxes[:, 4], 0, height)
    boxes[:, 5] = np.clip(boxes[:, 5], 0, width)

    return boxes


def detections2mask(detections, masks, img_reso, num_class=28):
    """
    Apply results of mask-rcnn (detections and masks) to mask result.

    detections: detected bounding boxes [z, y, x, d, h, w, category]
    masks: mask predictions correponding to each one of the detections config['mask_crop_size']
    img_reso: tuple with 3 elements, shape of the image or target resolution of the mask
    """
    D, H, W = img_reso
    mask = np.zeros((num_class, D, H, W))
    for i in range(len(detections)):
        z, y, x, d, h, w, cat = detections[i]

        cat = int(cat)
        z_start = max(0, int(np.floor(z - d / 2.)))
        y_start = max(0, int(np.floor(y - h / 2.)))
        x_start = max(0, int(np.floor(x - w / 2.)))
        z_end = min(D, int(np.ceil(z + d / 2.)))
        y_end = min(H, int(np.ceil(y + h / 2.)))
        x_end = min(W, int(np.ceil(x + w / 2.)))

        m = masks[i]
        D_c, H_c, W_c = m.shape
        zoomed_crop = zoom(m, 
                    (float(z_end - z_start) / D_c, float(y_end - y_start) / H_c, float(x_end - x_start) / W_c), 
                    order=2)
        mask[cat - 1][z_start:z_end, y_start:y_end, x_start:x_end] = (zoomed_crop > 0.5).astype(np.uint8)
    
    return mask


def crop_boxes2mask(crop_boxes, masks, img_reso, num_class=28):
    """
    Apply results of mask-rcnn (detections and masks) to mask result.

    crop_boxes: detected bounding boxes [z, y, x, d, h, w, category]
    masks: mask predictions correponding to each one of the detections config['mask_crop_size']
    img_reso: tuple with 3 elements, shape of the image or target resolution of the mask
    """
    D, H, W = img_reso
    mask = np.zeros((num_class, D, H, W))
    for i in range(len(crop_boxes)):
        z_start, y_start, x_start, z_end, y_end, x_end, cat = crop_boxes[i]

        cat = int(cat)

        m = masks[i]
        D_c, H_c, W_c = m.shape
        mask[cat - 1][z_start:z_end, y_start:y_end, x_start:x_end] = (m > 0.5).astype(np.uint8)
    
    return mask

def crop_boxes2mask_single(crop_boxes, masks, img_reso):
    """
    Apply results of mask-rcnn (detections and masks) to mask result.

    crop_boxes: detected bounding boxes [z, y, x, d, h, w, category]
    masks: mask predictions correponding to each one of the detections config['mask_crop_size']
    img_reso: tuple with 3 elements, shape of the image or target resolution of the mask
    """
    D, H, W = img_reso
    mask = np.zeros((D, H, W))
    for i in range(len(crop_boxes)):
        z_start, y_start, x_start, z_end, y_end, x_end, cat = crop_boxes[i]

        cat = int(cat)

        m = masks[i]
        D_c, H_c, W_c = m.shape
        mask[z_start:z_end, y_start:y_end, x_start:x_end][m > 0.5] = i + 1
    
    return mask


def masks2bboxes_masks(masks, border):
    """
    Generate bounding boxes from masks

    masks: [num_class, D, H, W]
    return: [z, y, x, class]
    """
    
    num_class, D, H, W = masks.shape
    bboxes = []
    truth_masks = []
    for i in range(num_class):
        mask = masks[i]
        if np.any(mask):
            zz, yy, xx = np.where(mask)
            bboxes.append([(zz.max() + zz.min()) / 2., 
                           (yy.max() + yy.min()) / 2., 
                           (xx.max() + xx.min()) / 2., 
                           zz.max() - zz.min() + 1 + border / 2, 
                           yy.max() - yy.min() + 1 + border, 
                           xx.max() - xx.min() + 1 + border, 
                           i + 1])
            truth_masks.append(mask)

    return bboxes, truth_masks


def masks2bboxes_masks_one(masks, border):
    "input mask shape: (128,128,128)"
    """
    Generate bounding boxes from masks

    masks: [num_class, D, H, W]
    return: [z, y, x, class]
    """
#     print "***inside function****",masks.shape
    D, H, W = masks.shape
    instance_nums = [num for num in np.unique(masks) if num]
#     print "***instance_nums***:",instance_nums
    bboxes = []
    truth_masks = []
    for i in instance_nums:
        mask = (masks == i)
        if np.any(mask):
            zz, yy, xx = np.where(mask)
            bboxes.append([(zz.max() + zz.min()) / 2.,
                           (yy.max() + yy.min()) / 2.,
                           (xx.max() + xx.min()) / 2.,
                           zz.max() - zz.min() + 1 + border,
                           yy.max() - yy.min() + 1 + border,
                           xx.max() - xx.min() + 1 + border,
                           1])
            truth_masks.append(mask)

    return bboxes, truth_masks



def extend_bbox(bboxes, extend_border):
    """
        input bbox shape: (Z, Y, X, D)
        return bbox shape: (Z, Y, X, D+board, D+board, D+board)
    """
    original_bbox = bboxes
    original_bbox[:,-1]+=extend_border
    D = np.array([original_bbox[:,-1]]).reshape(-1,1)
    return_bboxes = np.concatenate((original_bbox,D,D),axis=1)
    
    return return_bboxes



def bboxes_masks2masks(coord_boxes, masks, labels, reso, num_class=len(config['roi_names'])):
    """
    Generate masks from bounding boxes

    masks: [num_class, D, H, W]
    return: [z, y, x, class]
    """
    D, H, W = reso
    mask = np.zeros((num_class, D, H, W), dtype=np.uint8)
    for i in range(len(labels)):
        cat = int(labels[i]) - 1
        z_start, y_start, x_start, z_end, y_end, x_end = coord_boxes[i]
#         print coord_boxes[i]
#         print masks[i].shape
        mask[cat][z_start:z_end, y_start:y_end, x_start:x_end] = masks[i].astype(np.uint8)

    return mask


def get_contours_from_masks(masks):
    """
    Generate contours from masks by going through each organ slice by slice
    
    masks: [num_class, D, H, W]
    return: contours of shape [num_class, D, H, W] for each organ
    """
    contours = np.zeros(masks.shape, dtype=np.uint8)
    
    # Iterate all organs/channels
    for i, mask in enumerate(masks):
        # For each organ, Iterate all slices
        for j, s in enumerate(mask):
            c = np.zeros(s.shape)
            pts = measure.find_contours(s, 0)

            if pts:
                # There is contour in the image
                pts = np.concatenate(pts).astype(np.int32)
                for point in pts:
                    c[point[0], point[1]] = 1

            contours[i][j] = c
            
    return contours


def merge_contours(contours):
    """
    Merge contours for each organ into one ndimage, overlapped pixels will
    be override by the later class value
    
    contours: [num_class, D, H, W]
    return: merged contour of shape [D, H, W]
    """
    num_class, D, H, W = contours.shape
    merged_contours = np.zeros((D, H, W), dtype=np.uint8)
    for i in range(num_class):
        merged_contours[contours[i] > 0] = i + 1
    
    return merged_contours


def merge_masks(masks):
    """
    Merge masks for each organ into one ndimage, overlapped pixels will
    be override by the later class value
    
    contours: [num_class, D, H, W]
    return: merged contour of shape [D, H, W]
    """
    num_class, D, H, W = masks.shape
    merged_masks = np.zeros((D, H, W), dtype=np.uint8)
    for i in range(num_class):
        merged_masks[masks[i] > 0] = i + 1
    
    return merged_masks


def dice_score(y_pred, y_true, num_class=len(config['roi_names']), decimal=4):
    res = []
    for i in range(num_class):
        target = y_true == i
        pred = y_pred == i
        if target.sum():
            score = 2 * (target * pred).sum() / float((target.sum() + pred.sum()))
            res.append(round(score, decimal))
        else:
            res.append(None)

    return res


def dice_score_seperate(y_pred, y_true, num_class=len(config['roi_names']), decimal=4):
    res = []
    for i in range(num_class):
        target = y_true[i]
        pred = y_pred[i]
        if target.sum():
            score = 2 * (target * pred).sum() / float((target.sum() + pred.sum()))
            res.append(round(score, decimal))
        else:
            res.append(None)

    return res


def hausdorff_distance(y_pred, y_true, spacing=[1., 1., 1.], percent=0.95, num_class=len(config['roi_names']), decimal=4):
    """
    calculate the 95% (by default) hausdorff distance between the contour of prediction and ground truth
    """
    res = []

    for i in range(num_class):
        target = y_true[i]
        pred = y_pred[i]

        if target.sum() and pred.sum():
            a_pts = np.where(target)
            b_pts = np.where(pred)
            a_pts = np.array(a_pts).T * np.array(spacing)
            b_pts = np.array(b_pts).T * np.array(spacing)

            dists = cdist(a_pts, b_pts)
            a = np.min(dists, 1)
            b = np.min(dists, 0)
            a.sort()
            b.sort()

            a_max = a[int(percent * len(a)) - 1]
            b_max = b[int(percent * len(b)) - 1]

            res.append(round(max(a_max, b_max), decimal))
        else:
            res.append(None)
    return res


def pad2factor(image, factor=16, pad_value=0):
    depth, height, width = image.shape
    d = int(math.ceil(depth / float(factor))) * factor
    h = int(math.ceil(height / float(factor))) * factor
    w = int(math.ceil(width / float(factor))) * factor

    pad = []
    pad.append([0, d - depth])
    pad.append([0, h - height])
    pad.append([0, w - width])

    image = np.pad(image, pad, 'constant', constant_values=pad_value)

    return image


def normalize(img):
    maximum = img.max()
    minimum = img.min()

    # 0 ~ 1
    img = (img - minimum) / max(1, (maximum - minimum))
    
    # -1 ~ 1
    img = img * 2 - 1
    return img

def annotation2multi_mask(mask):
    multi_mask = np.zeros(mask[mask.keys()[0]].shape)
    for i, roi in enumerate(config['roi_names']):
        if roi in mask:
            multi_mask[mask[roi] > 0] = i + 1

    return multi_mask

def annotation2masks(mask):
    D, H, W = mask[mask.keys()[0]].shape
    masks = np.zeros([len(config['roi_names']), D, H, W])
    for i, roi in enumerate(config['roi_names']):
        if roi in mask:
            masks[i][mask[roi] > 0] = 1

    return masks

def multi_mask2onehot(mask):
    D, H, W = mask.shape
    onehot_mask = np.zeros((len(config['roi_names']) + 1, D, H, W))
    for i in range(len(config['roi_names']) + 1):
        onehot_mask[i][mask == i] = 1

    return onehot_mask

def onehot2multi_mask(onehot):
    num_class, D, H, W = onehot.shape
    multi_mask = np.zeros((D, H, W))

    for i in range(1, num_class):
        multi_mask[onehot[i] > 0] = i

    return multi_mask

def load_dicom_image(foldername):
    reader = sitk.ImageSeriesReader()
    dicom_names = reader.GetGDCMSeriesFileNames(foldername)
    reader.SetFileNames(dicom_names)
    itkimage = reader.Execute()
    numpyImage = sitk.GetArrayFromImage(itkimage)

    numpyOrigin = np.array(list(reversed(itkimage.GetOrigin())))
    numpySpacing = np.array(list(reversed(itkimage.GetSpacing())))

    return numpyImage, numpyOrigin, numpySpacing


def truncate_HU_uint8(img):
    """Truncate HU range and convert to uint8."""

    HU_range = np.array([-1200., 600.])
    new_img = (img - HU_range[0]) / (HU_range[1] - HU_range[0])
    new_img[new_img < 0] = 0
    new_img[new_img > 1] = 1
    new_img = (new_img * 255).astype('uint8')
    return new_img

def py_nms():
    raise NotImplementedError("py_nms not implemented")

def py_box_overlap():
    raise NotImplementedError("py_box_overlap not implemented")

def pad2factor(image, factor=32, pad_value=0):
    # print('Inside pad2factor')
    # print('-'*40)
    depth, height, width = image.shape
    # print('depth, height, width:', depth, height, width)
    d = int(math.ceil(depth / float(factor))) * factor
    # print("d = int(math.ceil({}/{})) * {} = {}".format(depth, factor, factor, d))
    h = int(math.ceil(height / float(factor))) * factor
    # print("h = int(math.ceil({}/{})) * {} = {}".format(height, factor, factor, h))
    w = int(math.ceil(width / float(factor))) * factor
    # print("w = int(math.ceil({}/{})) * {} = {}".format(width, factor, factor, w))

    pad = []
    pad.append([0, d - depth])
    pad.append([0, h - height])
    pad.append([0, w - width])
    # print('adding pad: [{}-{}, {}-{}, {}-{}] = [{}, {}, {}]'.format(d, depth, h, height, w, width, d-depth, h-height, w-width))

    # pad = []
    # pad.append([0, w - width + height-depth])
    # pad.append([0, h - height])
    # pad.append([0, w - width])

    image = np.pad(image, pad, 'constant', constant_values=pad_value)

    return image


from setting import parse_opts 
from datasets.brains18 import BrainS18Dataset
from datasets.lungCT import LIDC_CT
from model import generate_model
import torch
import numpy as np
from torch.utils.data import DataLoader
import torch.nn.functional as F
from scipy import ndimage
from torch import nn
import nibabel as nib
import sys
import cv2
import os
import numpy as np
import nibabel
from build_unet_model import UNet


def getIoU(y_true, y_pred):
    y_true_f = y_true.flatten()
    y_pred_f = y_pred.flatten()
    intersection = np.logical_and(y_true_f, y_pred_f).sum()
    union = np.logical_or(y_true_f, y_pred_f).sum()
    return (intersection + 1) * 1. / (union + 1)


def rescale_img_to_rgb(img):
    img = (img - img.min()) / (img.max() - img.min())
    img *= 255.0
    img = img.astype(np.uint8)
    return img


def test(data_loader, model, img_names, sets):
    masks = []
    model.eval() # for testing 
    for batch_id, batch_data in enumerate(data_loader):
        # forward
        volume = batch_data
        if not sets.no_cuda:
            volume = volume.cuda()
        with torch.no_grad():
            probs = model(volume)
            probs = F.softmax(probs, dim=1)

        # resize mask to original size
        [batchsize, _, mask_d, mask_h, mask_w] = probs.shape
        data = nib.load(os.path.join(sets.data_root, img_names[batch_id]))
        data = data.get_data()
        [depth, height, width] = data.shape
        mask = probs[0]
        scale = [1, depth*1.0/mask_d, height*1.0/mask_h, width*1.0/mask_w]
        mask = ndimage.interpolation.zoom(mask, scale, order=1)
        mask = np.argmax(mask, axis=0)        
        masks.append(mask)
    return masks

def resize_data__(data):
    """
    Resize the data to the input size
    """ 
    [height, width, depth] = data.shape
    # scale = [self.input_D*1.0/depth, self.input_H*1.0/height, self.input_W*1.0/width] 
    scale = [64*1.0/height, 64*1.0/width, 32*1.0/depth]  
    data = ndimage.interpolation.zoom(data, scale, order=0)
    return data



if __name__ == '__main__':
    # settting
    sets = parse_opts()

    # os.environ["CUDA_VISIBLE_DEVICES"]=str(sets.gpu_id[0])   
    # model = UNet(in_nc=1, out_nc=1, nf=64).cuda()
    # model = nn.DataParallel(model)

    model, _ = generate_model(sets)

    # getting weights
    path_to_weight = './trails_ResNet-34/models/resnet_34_epoch_187_best_weight_0.372868.pth.tar'   
    checkpoint = torch.load(path_to_weight)
    model.load_state_dict(checkpoint['state_dict'])
    print('Model weights loaded!')

    # load image and mask
    img_path = nibabel.load('./data/data_lidc/segmentation/patches/images/LIDC-IDRI-0014-265-nod-345-71.nii.gz')
    img = img_path.get_fdata()
    img = rescale_img_to_rgb(img)

    mask_path = nibabel.load('./data/data_lidc/segmentation/patches/labels/LIDC-IDRI-0014-265-nod-345-71.nii.gz')
    mask = mask_path.get_fdata()

    img = resize_data__(img)
    cv2.imwrite('LIDC-IDRI-0014-265-nod-345-71.png', img[:,:,16])
    mask = resize_data__(mask)

    img = (img - img.mean()) / img.std()
    img = torch.from_numpy(np.expand_dims(np.expand_dims(img.transpose(2,0,1), axis=0), axis=0)).type(torch.FloatTensor)
    label_masks = torch.from_numpy(np.expand_dims(np.expand_dims(mask.transpose(2,0,1), axis=0), axis=0)).type(torch.FloatTensor)

    print('img shape:', img.shape)
    print('GT mask shape:', label_masks.shape)
    
    with torch.no_grad():
        model.eval()
        img = img.cuda()
        out_masks = model(img)

        output_prob = torch.sigmoid(out_masks).detach().cpu().numpy()
        output_prob_thresh = (output_prob > 0.5) * 1
        print('Out mask shape:', output_prob_thresh.shape)
        print('Pred unique:', np.unique(output_prob_thresh))

        # resize GT label
        [n, _, d, h, w] = out_masks.shape
        new_label_masks = np.zeros([n, _, d, h, w])
        for label_id in range(n):
            label_mask = label_masks[label_id]
            [ori_c, ori_d, ori_h, ori_w] = label_mask.shape 
            label_mask = np.reshape(label_mask, [ori_d, ori_h, ori_w])
            scale = [d*1.0/ori_d, h*1.0/ori_h, w*1.0/ori_w]
            label_mask = ndimage.interpolation.zoom(label_mask, scale, order=0)
            new_label_masks[label_id, 0] = label_mask
        new_label_masks = torch.tensor(new_label_masks).type(torch.FloatTensor).numpy()
        
        print('Downsampled mask unique:', np.unique(new_label_masks))
        iou = getIoU(new_label_masks, output_prob_thresh)
        print('IoU is:', iou)
        
        # print('Resized GT label shape:', new_label_masks.shape)
        # upsample predictions
        [n, _, d, h, w] = label_masks.shape
        [batch, ori_c, ori_d, ori_h, ori_w] = output_prob_thresh.shape
        scale = [d*1.0/ori_d, h*1.0/ori_h, w*1.0/ori_w]
        output_prob_thresh = ndimage.interpolation.zoom(output_prob_thresh[0,0,:,:,:], scale, order=0)

        # iou = getIoU(label_masks, output_prob_thresh)
        # print('IoU is:', iou)

# middle_pd = output_prob_thresh[0,0,16,:,:]
middle_pd = output_prob_thresh[16,:,:]
middle_pd = middle_pd*255.0
middle_pd = middle_pd.astype(np.uint8)
cv2.imwrite('LIDC-IDRI-0014-265-nod-345-71_prediction.png', middle_pd)

# # gt_middle = mask[:,:, 16]
gt_middle = label_masks.numpy()[0, 0, 16, :,:]
gt_middle = gt_middle*255.0
gt_middle = gt_middle.astype(np.uint8)
cv2.imwrite('LIDC-IDRI-0014-265-nod-345-71_gt.png', gt_middle)



'''create dataset and dataloader'''
import logging
import torch.utils.data


"""
Create dataloader given the dataset object and other parameter dictionary
"""
def create_dataloader(dataset, dataset_opt):
    # phase = dataset_opt['phase']
    # # return dataloader depending on phase, either 'train' or 'test'
    # if phase == 'train':
    #     return torch.utils.data.DataLoader(
    #         dataset,
    #         batch_size=dataset_opt['batch_size'],
    #         shuffle=dataset_opt['use_shuffle'],
    #         num_workers=dataset_opt['n_workers'],
    #         drop_last=True,
    #         pin_memory=True)
    # else:
    return torch.utils.data.DataLoader(
        dataset, batch_size=1, shuffle=False, num_workers=1)


"""
Create Dataset object for dataloader
"""
def create_dataset(dataset_opt):
    # import dataset class depending on 'data_type'
    if dataset_opt['image_type'] == 'h5':
        from .h5Dataset_with_label import h5Dataset as D
    elif dataset_opt['image_type'] == 'dicom':
        # from .DcmDataset import dcmDataset as D
        from .nlstDataset import dcmDataset as D
    elif dataset_opt['image_type'] == 'nii':
        from .niiDataset_with_label import NIIDataset as D

    # # load uids in this phase    
    # with open(dataset_opt['uids_path'], 'r') as f:
    #     lines = f.readlines() # ['1d7425c7412df778d26c027dc5f46a38\n', ....]
    #     # remove \n from each element in 'lines' list
    #     dataset_opt['uids'] = [l.rstrip() for l in lines]
    dataset = D(dataset_opt) # create dataobject
    logger = logging.getLogger('base')
    logger.info('Dataset [{:s} - {:s}] is created.'.format(dataset.__class__.__name__,
                                                           dataset_opt['source']))
    return dataset



from __future__ import absolute_import
import sys
sys.path.append('..')
sys.path.append('.')
import numpy as np
import torch
from torch import nn
import os
from collections import OrderedDict
from torch.autograd import Variable
import itertools
from .base_model import BaseModel
from scipy.ndimage import zoom
import fractions
import functools
import skimage.transform
from IPython import embed
from . import networks_basic as networks
from ..util import util

class DistModel(BaseModel):
    """
    Returns model name
    """
    def name(self):
        return self.model_name

    """
    Initializes pdst model (e.g. PNetLin), loads weights and sets up for either training or evaluation
    """
    def initialize(self, model='net-lin', net='alex', pnet_rand=False, pnet_tune=False, model_path=None, colorspace='Lab', use_gpu=True, printNet=True, spatial=False, spatial_shape=None, spatial_order=1, spatial_factor=None, is_train=False, lr=.0001, beta1=0.5, version='0.1'):
        """
        INPUTS
            model - ['net-lin'] for linearly calibrated network
                    ['net'] for off-the-shelf network
                    ['L2'] for L2 distance in Lab colorspace
                    ['SSIM'] for ssim in RGB colorspace
            net - ['squeeze','alex','vgg']
            model_path - if None, will look in weights/[NET_NAME].pth
            colorspace - ['Lab','RGB'] colorspace to use for L2 and SSIM
            use_gpu - bool - whether or not to use a GPU
            printNet - bool - whether or not to print network architecture out
            spatial - bool - whether to output an array containing varying distances across spatial dimensions
            spatial_shape - if given, output spatial shape. if None then spatial shape is determined automatically via spatial_factor (see below).
            spatial_factor - if given, specifies upsampling factor relative to the largest spatial extent of a convolutional layer. if None then resized to size of input images.
            spatial_order - spline order of filter for upsampling in spatial mode, by default 1 (bilinear).
            is_train - bool - [True] for training mode
            lr - float - initial learning rate
            beta1 - float - initial momentum term for adam
            version - 0.1 for latest, 0.0 was original
        """
        BaseModel.initialize(self, use_gpu=use_gpu)

        self.model = model
        self.net = net
        self.use_gpu = use_gpu
        self.is_train = is_train
        self.spatial = spatial
        self.spatial_shape = spatial_shape
        self.spatial_order = spatial_order
        self.spatial_factor = spatial_factor

        print("dist_model initialize function called")
        print("model:", self.model)
        print("net:", self.net)
        print("use gpu:", self.use_gpu)
        print("is_train:", self.is_train)
        print("spatial:", self.spatial)
        print("spatial shape:", self.spatial_shape)
        print("spatial order:", self.spatial_order)
        print("spatial factor:", self.spatial_factor)

        # model name: net-lin [vgg]
        self.model_name = '%s [%s]'%(model, net)
        print("model name:", self.model_name)

        if(self.model == 'net-lin'): # pretrained net + linear layer

            # use_gpu=True, pnet_rand=False, pnet_tune=False, pnet_type=vgg, spatial=False, version=0.1
            self.net = networks.PNetLin(use_gpu=use_gpu,pnet_rand=pnet_rand, pnet_tune=pnet_tune, pnet_type=net,use_dropout=True,spatial=spatial,version=version)
            kw = {}

            if not use_gpu:
                print("map_location will be cpu")
                kw['map_location'] = 'cpu'

            # since 'model_path' is None, inspect for vgg16 weights path based on weights version (0.1)
            if(model_path is None):
                import inspect
                # model_path = './PerceptualSimilarity/weights/v%s/%s.pth'%(version,net)
                model_path = os.path.abspath(os.path.join(inspect.getfile(self.initialize), '..', '..', 'weights/v%s/%s.pth'%(version,net)))

            # 'is_train' is False; not False is True, so we load the vgg16 weights to the model
            if(not is_train):
                print('Loading model from: %s'%model_path)
                self.net.load_state_dict(torch.load(model_path, **kw))

        elif(self.model=='net'): # pretrained network
            assert not self.spatial, 'spatial argument not supported yet for uncalibrated networks'
            self.net = networks.PNet(use_gpu=use_gpu,pnet_type=net)
            self.is_fake_net = True
        elif(self.model in ['L2','l2']):
            self.net = networks.L2(use_gpu=use_gpu,colorspace=colorspace) # not really a network, only for testing
            self.model_name = 'L2'
        elif(self.model in ['DSSIM','dssim','SSIM','ssim']):
            self.net = networks.DSSIM(use_gpu=use_gpu,colorspace=colorspace)
            self.model_name = 'SSIM'
        else:
            raise ValueError("Model [%s] not recognized." % self.model)

        # get all the parameters from the network
        self.parameters = list(self.net.parameters())

        # if 'self.is_train' is True, we add extra network (self.rankLoss) on top to go from 
        # distances (d0, d1)=> predicted human judgment (h*), and then add 'self.rankLoss' model
        # parameters to existing 'self.parameters' list, and also set up optimizer
        if self.is_train: # training mode
            # extra network on top to go from distances (d0,d1) 
            self.rankLoss = networks.BCERankingLoss(use_gpu=use_gpu)
            self.parameters+=self.rankLoss.parameters
            self.lr = lr
            self.old_lr = lr
            self.optimizer_net = torch.optim.Adam(self.parameters, lr=lr, betas=(beta1, 0.999))
        else: # put model on evaluation mode
            self.net.eval()

        # print the pdist model, if specified
        if(printNet):
            print('---------- Pdist Networks initialized -------------')
            networks.print_network(self.net)
            print('-----------------------------------------------')

    def forward_pair(self,in1,in2,retPerLayer=False):
        if(retPerLayer):
            return self.net.forward(in1,in2, retPerLayer=True)
        else:
            return self.net.forward(in1,in2)

    def forward(self, in0, in1, retNumpy=True):
        ''' Function computes the distance between image patches in0 and in1
        INPUTS
            in0, in1 - torch.Tensor object of shape Nx3xXxY - image patch scaled to [-1,1]
            retNumpy - [False] to return as torch.Tensor, [True] to return as numpy array
        OUTPUT
            computed distances between in0 and in1
        '''
        self.input_ref = in0
        self.input_p0 = in1

        if(self.use_gpu):
            self.input_ref = self.input_ref.cuda()
            self.input_p0 = self.input_p0.cuda()

        self.var_ref = Variable(self.input_ref,requires_grad=True)
        self.var_p0 = Variable(self.input_p0,requires_grad=True)

        self.d0 = self.forward_pair(self.var_ref, self.var_p0)
        self.loss_total = self.d0

        def convert_output(d0):
            if(retNumpy):
                ans = d0.cpu().data.numpy()
                if not self.spatial:
                    ans = ans.flatten()
                else:
                    assert(ans.shape[0] == 1 and len(ans.shape) == 4)
                    return ans[0,...].transpose([1, 2, 0])                  # Reshape to usual numpy image format: (height, width, channels)
                return ans
            else:
                return d0

        if self.spatial:
            L = [convert_output(x) for x in self.d0]
            spatial_shape = self.spatial_shape
            if spatial_shape is None:
                if(self.spatial_factor is None):
                    spatial_shape = (in0.size()[2],in0.size()[3])
                else:
                    spatial_shape = (max([x.shape[0] for x in L])*self.spatial_factor, max([x.shape[1] for x in L])*self.spatial_factor)
            
            L = [skimage.transform.resize(x, spatial_shape, order=self.spatial_order, mode='edge') for x in L]
            
            L = np.mean(np.concatenate(L, 2) * len(L), 2)
            return L
        else:
            return convert_output(self.d0)

    # ***** TRAINING FUNCTIONS *****
    def optimize_parameters(self):
        self.forward_train()
        self.optimizer_net.zero_grad()
        self.backward_train()
        self.optimizer_net.step()
        self.clamp_weights()

    def clamp_weights(self):
        for module in self.net.modules():
            if(hasattr(module, 'weight') and module.kernel_size==(1,1)):
                module.weight.data = torch.clamp(module.weight.data,min=0)

    def set_input(self, data):
        self.input_ref = data['ref']
        self.input_p0 = data['p0']
        self.input_p1 = data['p1']
        self.input_judge = data['judge']

        if(self.use_gpu):
            self.input_ref = self.input_ref.cuda()
            self.input_p0 = self.input_p0.cuda()
            self.input_p1 = self.input_p1.cuda()
            self.input_judge = self.input_judge.cuda()

        self.var_ref = Variable(self.input_ref,requires_grad=True)
        self.var_p0 = Variable(self.input_p0,requires_grad=True)
        self.var_p1 = Variable(self.input_p1,requires_grad=True)

    def forward_train(self): # run forward pass
        self.d0 = self.forward_pair(self.var_ref, self.var_p0)
        self.d1 = self.forward_pair(self.var_ref, self.var_p1)
        self.acc_r = self.compute_accuracy(self.d0,self.d1,self.input_judge)

        # var_judge
        self.var_judge = Variable(1.*self.input_judge).view(self.d0.size())

        self.loss_total = self.rankLoss.forward(self.d0, self.d1, self.var_judge*2.-1.)
        return self.loss_total

    def backward_train(self):
        torch.mean(self.loss_total).backward()

    def compute_accuracy(self,d0,d1,judge):
        ''' d0, d1 are Variables, judge is a Tensor '''
        d1_lt_d0 = (d1<d0).cpu().data.numpy().flatten()
        judge_per = judge.cpu().numpy().flatten()
        return d1_lt_d0*judge_per + (1-d1_lt_d0)*(1-judge_per)

    def get_current_errors(self):
        retDict = OrderedDict([('loss_total', self.loss_total.data.cpu().numpy()),
                            ('acc_r', self.acc_r)])

        for key in retDict.keys():
            retDict[key] = np.mean(retDict[key])

        return retDict

    def get_current_visuals(self):
        zoom_factor = 256/self.var_ref.data.size()[2]

        ref_img = util.tensor2im(self.var_ref.data)
        p0_img = util.tensor2im(self.var_p0.data)
        p1_img = util.tensor2im(self.var_p1.data)

        ref_img_vis = zoom(ref_img,[zoom_factor, zoom_factor, 1],order=0)
        p0_img_vis = zoom(p0_img,[zoom_factor, zoom_factor, 1],order=0)
        p1_img_vis = zoom(p1_img,[zoom_factor, zoom_factor, 1],order=0)

        return OrderedDict([('ref', ref_img_vis),
                            ('p0', p0_img_vis),
                            ('p1', p1_img_vis)])

    def save(self, path, label):
        self.save_network(self.net, path, '', label)
        self.save_network(self.rankLoss.net, path, 'rank', label)

    def update_learning_rate(self,nepoch_decay):
        lrd = self.lr / nepoch_decay
        lr = self.old_lr - lrd

        for param_group in self.optimizer_net.param_groups:
            param_group['lr'] = lr

        print('update lr [%s] decay: %f -> %f' % (type,self.old_lr, lr))
        self.old_lr = lr



def score_2afc_dataset(data_loader,func):
    ''' Function computes Two Alternative Forced Choice (2AFC) score using
        distance function 'func' in dataset 'data_loader'
    INPUTS
        data_loader - CustomDatasetDataLoader object - contains a TwoAFCDataset inside
        func - callable distance function - calling d=func(in0,in1) should take 2
            pytorch tensors with shape Nx3xXxY, and return numpy array of length N
    OUTPUTS
        [0] - 2AFC score in [0,1], fraction of time func agrees with human evaluators
        [1] - dictionary with following elements
            d0s,d1s - N arrays containing distances between reference patch to perturbed patches 
            gts - N array in [0,1], preferred patch selected by human evaluators
                (closer to "0" for left patch p0, "1" for right patch p1,
                "0.6" means 60pct people preferred right patch, 40pct preferred left)
            scores - N array in [0,1], corresponding to what percentage function agreed with humans
    CONSTS
        N - number of test triplets in data_loader
    '''

    d0s = []
    d1s = []
    gts = []

    # bar = pb.ProgressBar(max_value=data_loader.load_data().__len__())
    for (i,data) in enumerate(data_loader.load_data()):
        d0s+=func(data['ref'],data['p0']).tolist()
        d1s+=func(data['ref'],data['p1']).tolist()
        gts+=data['judge'].cpu().numpy().flatten().tolist()
        # bar.update(i)

    d0s = np.array(d0s)
    d1s = np.array(d1s)
    gts = np.array(gts)
    scores = (d0s<d1s)*(1.-gts) + (d1s<d0s)*gts + (d1s==d0s)*.5

    return(np.mean(scores), dict(d0s=d0s,d1s=d1s,gts=gts,scores=scores))

def score_jnd_dataset(data_loader,func):
    ''' Function computes JND score using distance function 'func' in dataset 'data_loader'
    INPUTS
        data_loader - CustomDatasetDataLoader object - contains a JNDDataset inside
        func - callable distance function - calling d=func(in0,in1) should take 2
            pytorch tensors with shape Nx3xXxY, and return numpy array of length N
    OUTPUTS
        [0] - JND score in [0,1], mAP score (area under precision-recall curve)
        [1] - dictionary with following elements
            ds - N array containing distances between two patches shown to human evaluator
            sames - N array containing fraction of people who thought the two patches were identical
    CONSTS
        N - number of test triplets in data_loader
    '''

    ds = []
    gts = []

    # bar = pb.ProgressBar(max_value=data_loader.load_data().__len__())
    for (i,data) in enumerate(data_loader.load_data()):
        ds+=func(data['p0'],data['p1']).tolist()
        gts+=data['same'].cpu().numpy().flatten().tolist()
        # bar.update(i)

    sames = np.array(gts)
    ds = np.array(ds)

    sorted_inds = np.argsort(ds)
    ds_sorted = ds[sorted_inds]
    sames_sorted = sames[sorted_inds]

    TPs = np.cumsum(sames_sorted)
    FPs = np.cumsum(1-sames_sorted)
    FNs = np.sum(sames_sorted)-TPs

    precs = TPs/(TPs+FPs)
    recs = TPs/(TPs+FNs)
    score = util.voc_ap(recs,precs)

    return(score, dict(ds=ds,sames=sames))



import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Variable
import math
from functools import partial
from lung_analysis_pipeline.NoduleDetect_SANet.net import cgnl

__all__ = [
    'ResNet', 'resnet10', 'resnet18', 'resnet34', 'resnet50', 'resnet101',
    'resnet152', 'resnet200'
]


def conv3x3x3(in_planes, out_planes, stride=1):
    # 3x3x3 convolution with padding
    return nn.Conv3d(
        in_planes,
        out_planes,
        kernel_size=3,
        stride=stride,
        padding=1,
        bias=False)


def downsample_basic_block(x, planes, stride):
    out = F.avg_pool3d(x, kernel_size=1, stride=stride)
    zero_pads = torch.Tensor(
        out.size(0), planes - out.size(1), out.size(2), out.size(3),
        out.size(4)).zero_()
    if isinstance(out.data, torch.cuda.FloatTensor):
        zero_pads = zero_pads.cuda()

    out = Variable(torch.cat([out.data, zero_pads], dim=1))

    return out


class BasicBlock(nn.Module):
    expansion = 1

    def __init__(self, inplanes, planes, stride=1, downsample=None):
        super(BasicBlock, self).__init__()
        self.conv1 = conv3x3x3(inplanes, planes, stride)
        self.bn1 = nn.BatchNorm3d(planes)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = conv3x3x3(planes, planes)
        self.bn2 = nn.BatchNorm3d(planes)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.relu(out)

        return out


class Bottleneck(nn.Module):
    expansion = 1

    def __init__(self, inplanes, planes, stride=1, downsample=None):
        super(Bottleneck, self).__init__()
        self.conv1 = nn.Conv3d(inplanes, planes, kernel_size=1, bias=False)
        self.bn1 = nn.BatchNorm3d(planes)
        self.conv2 = nn.Conv3d(
            planes, planes, kernel_size=3, stride=stride, padding=1, bias=False)
        self.bn2 = nn.BatchNorm3d(planes)
        self.conv3 = nn.Conv3d(planes, planes, kernel_size=1, bias=False)
        self.bn3 = nn.BatchNorm3d(planes)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.relu(out)

        return out


class ResNet(nn.Module):

    def __init__(self,
                 block,
                 layers,
                 sample_size=None,
                 sample_duration=None,
                 shortcut_type='B',
                 num_classes=400):
        self.inplanes = 64
        super(ResNet, self).__init__()
        self.conv1 = nn.Conv3d(
            1,
            64,
            kernel_size=7,
            stride=(2, 2, 2),
            padding=(3, 3, 3),
            bias=False)
        self.bn1 = nn.BatchNorm3d(64)
        self.relu = nn.ReLU(inplace=True)
        self.maxpool = nn.MaxPool3d(kernel_size=(3, 3, 3), stride=2, padding=1)

        # self.preBlock = nn.Sequential(
        #         nn.Conv3d(1, 64, kernel_size = 3, padding = 1, stride=2),
        #         nn.BatchNorm3d(32),
        #         nn.ReLU(inplace = True),
        #         nn.Conv3d(64, 64, kernel_size = 3, padding = 1),
        #         nn.BatchNorm3d(64),
        #         nn.ReLU(inplace = True))

        self.layer1 = self._make_layer(block, 64, layers[0], shortcut_type)
        self.layer2 = self._make_layer(
            block, 64, layers[1], shortcut_type, stride=2)
        self.layer3 = self._make_layer(
            block, 64, layers[2], shortcut_type, stride=2)
        self.layer4 = self._make_layer(
            block, 64, layers[3], shortcut_type, stride=2)
        # last_duration = int(math.ceil(sample_duration / 16))
        # last_size = int(math.ceil(sample_size / 32))
        # self.avgpool = nn.AvgPool3d(
        #     (last_duration, last_size, last_size), stride=1)
        # self.fc = nn.Linear(512 * block.expansion, num_classes)

        for m in self.modules():
            if isinstance(m, nn.Conv3d):
                # m.weight = nn.init.kaiming_normal(m.weight, mode='fan_out')
                nn.init.kaiming_normal_(m.weight,
                                        mode='fan_out',
                                        nonlinearity='relu')
            elif isinstance(m, nn.BatchNorm3d):
                # m.weight.data.fill_(1)
                # m.bias.data.zero_()
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

    def _make_layer(self, block, planes, blocks, shortcut_type, stride=1):
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            if shortcut_type == 'A':
                downsample = partial(
                    downsample_basic_block,
                    planes=planes * block.expansion,
                    stride=stride)
            else:
                downsample = nn.Sequential(
                    nn.Conv3d(
                        self.inplanes,
                        planes * block.expansion,
                        kernel_size=1,
                        stride=stride,
                        bias=False), nn.BatchNorm3d(planes * block.expansion))

        layers = []
        layers.append(block(self.inplanes, planes, stride, downsample))
        self.inplanes = planes * block.expansion
        for i in range(1, blocks):
            if (i == 5 and blocks == 6) or (i == 3 and blocks == 6) or (i == 1 and blocks == 6) or \
                    (i == 3 and blocks == 4) or (i == 1 and blocks == 4):
                layers.append(cgnl.SpatialCGNL(
                    self.inplanes,
                    int(self.inplanes / 2),
                    use_scale=False,
                    groups=4))

            layers.append(block(self.inplanes, planes))

        return nn.Sequential(*layers)

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        # x = self.maxpool(x)

        out1 = self.layer1(x)
        out2 = self.layer2(out1)
        out3 = self.layer3(out2)
        out4 = self.layer4(out3)

        # x = self.avgpool(x)

        # x = x.view(x.size(0), -1)
        # x = self.fc(x)

        return x, out1, out2, out3, out4


def get_fine_tuning_parameters(model, ft_portion):
    if ft_portion == "complete":
        return model.parameters()

    elif ft_portion == "last_layer":
        ft_module_names = []
        ft_module_names.append('fc')

        parameters = []
        for k, v in model.named_parameters():
            for ft_module in ft_module_names:
                if ft_module in k:
                    parameters.append({'params': v})
                    break
            else:
                parameters.append({'params': v, 'lr': 0.0})
        return parameters

    else:
        raise ValueError("Unsupported ft_portion: 'complete' or 'last_layer' expected")


def resnet10(**kwargs):
    """Constructs a ResNet-18 model.
    """
    model = ResNet(BasicBlock, [1, 1, 1, 1], **kwargs)
    return model


def resnet18(**kwargs):
    """Constructs a ResNet-18 model.
    """
    model = ResNet(BasicBlock, [2, 2, 2, 2], **kwargs)
    return model


def resnet34(**kwargs):
    """Constructs a ResNet-34 model.
    """
    model = ResNet(BasicBlock, [3, 4, 6, 3], **kwargs)
    return model


def resnet50(**kwargs):
    """Constructs a ResNet-50 model.
    """
    model = ResNet(Bottleneck, [3, 4, 6, 3], **kwargs)
    return model


def resnet101(**kwargs):
    """Constructs a ResNet-101 model.
    """
    model = ResNet(Bottleneck, [3, 4, 23, 3], **kwargs)
    return model


def resnet152(**kwargs):
    """Constructs a ResNet-101 model.
    """
    model = ResNet(Bottleneck, [3, 8, 36, 3], **kwargs)
    return model


def resnet200(**kwargs):
    """Constructs a ResNet-101 model.
    """
    model = ResNet(Bottleneck, [3, 24, 36, 3], **kwargs)
    return model



import os
import torch
from torch.utils.ffi import create_extension


sources = ['src/overlap.c']
headers = ['src/overlap.h']
defines = []
with_cuda = False

if torch.cuda.is_available():
    print('Including CUDA code.')
    sources += ['src/overlap_cuda.c']
    headers += ['src/overlap_cuda.h']
    defines += [('WITH_CUDA', None)]
    with_cuda = True

this_file = os.path.dirname(os.path.realpath(__file__))
print(this_file)
extra_objects = ['src/overlap_kernel.cu.o']
extra_objects = [os.path.join(this_file, fname) for fname in extra_objects]

ffi = create_extension(
    'extension',
    headers       = headers,
    sources       = sources,
    define_macros = defines,
    relative_to   = __file__,
    with_cuda     = with_cuda,
    extra_objects = extra_objects
)

if __name__ == '__main__':
    ffi.build()